unit LazUnicode;

{$mode objfpc}{$H+}

// For testing the UTF16 version.
{$IFDEF UseUTF16}
{$ModeSwitch UnicodeStrings}   // Sets also FPC_UNICODESTRINGS ...
{$DEFINE FPC_UNICODESTRINGS}   // but Lazarus editor hi-lighting doesn't know it.
{$ENDIF}

interface

uses
  Classes, SysUtils,
  {$IFDEF FPC_UNICODESTRINGS}
  character, LazUTF16,
  {$ELSE}
  {$ENDIF}
  LazUTF8;

  {$IFDEF FPC_UNICODESTRINGS}
  // Used with ModeSwitch UnicodeStrings.
  function UnicodeCopy(const s: UnicodeString; StartCharIndex, CharCount: PtrInt): Unicodestring;
  function UTF16CharStart(P: PUnicodeChar; Len, CharIndex: PtrInt): PUnicodeChar;
  function UTF16Pos(const SearchForText, SearchInText: UnicodeString; StartPos: SizeInt = 1): PtrInt;
  {$ENDIF}

  // Helper functions for codepoints. They change behavior depending on ModeSwitch.
  function CodePointCopy(const s: string; StartCharIndex, CharCount: PtrInt): string;
  function CodePointLength(const s: string): PtrInt;
  function CodePointPos(const SearchForText, SearchInText: string; StartPos: SizeInt = 1): PtrInt;
  function CodePointToWinCP(const s: string): AnsiString;
  function WinCPToCodePoint(const s: AnsiString): string;
  function CodeUnitCount(p: PChar): integer;
  function IsCombining(const AChar: PChar): Boolean;

type
  // Base class for CodePoint and Character enumerators.
  TUnicodeEnumeratorBase = class
  private
    fCurrent: String;
    fCurrentPos, fEndPos: PChar;
    fCurrentCodeUnitCount: Integer;
  public
    constructor Create(const A: String);
    property Current: String read fCurrent;
    property CurrentCodeUnitCount: Integer read fCurrentCodeUnitCount;
  end;

  // Traverse Unicode codepoints. Uses UTF-8 or UTF-16 depending on $ModeSwitch.
  TCodePointEnumerator = class(TUnicodeEnumeratorBase)
  public
    function MoveNext: Boolean;
  end;

  // Traverse Unicode (user perceived) characters, including accented characters
  //  with combined codepoints. Uses UTF-8 or UTF-16 depending on $ModeSwitch.
  // Used for an operator in for-in loop.

  { TUnicodeCharacterEnumerator }

  TUnicodeCharacterEnumerator = class(TCodePointEnumerator)
  private
    fCurrentCodePointCount: Integer;
  public
    property CurrentCodePointCount: Integer read fCurrentCodePointCount;
    function MoveNext: Boolean;
  end;

  operator Enumerator(A: String): TUnicodeCharacterEnumerator;

implementation

{$IFDEF FPC_UNICODESTRINGS}
function UnicodeCopy(const s: UnicodeString; StartCharIndex, CharCount: PtrInt): Unicodestring;
// returns substring
var
  StartPos: PUnicodeChar;
  EndPos: PUnicodeChar;
  MaxBytes: PtrInt;
begin
  StartPos:=UTF16CharStart(PUnicodeChar(s),length(s),StartCharIndex-1);
  if StartPos=nil then
    Result:=''
  else begin
    MaxBytes:=PtrInt(PUnicodeChar(s)+length(s)-StartPos);
    EndPos:=UTF16CharStart(StartPos,MaxBytes,CharCount);
    if EndPos=nil then
      Result:=copy(s,StartPos-PUnicodeChar(s)+1,MaxBytes)
    else
      Result:=copy(s,StartPos-PUnicodeChar(s)+1,EndPos-StartPos);
  end;
end;

function UTF16CharStart(P: PUnicodeChar; Len, CharIndex: PtrInt): PUnicodeChar;
// Len is the length in words of P.
// CharIndex is the position of the desired UnicodeChar (starting at 0).
var
  CharLen: LongInt;
begin
  Result:=P;
  if Result=nil then Exit;
  while (CharIndex>0) and (Len>0) do
  begin
    CharLen:=UTF16CharacterLength(Result);
    dec(Len,CharLen);
    dec(CharIndex);
    inc(Result,CharLen);
  end;
  if (CharIndex<>0) or (Len<0) then
    Result:=nil;
end;

// Helper used by UTF16Pos
function UTF16PosP(SearchForText: PUnicodeChar; SearchForTextLen: SizeInt;
  SearchInText: PUnicodeChar; SearchInTextLen: SizeInt): PUnicodeChar;
// returns the position where SearchInText starts in SearchForText
// returns nil if not found
var
  p: SizeInt;
begin
  Result:=nil;
  if (SearchForText=nil) or (SearchForTextLen=0) or (SearchInText=nil) then
    exit;
  while SearchInTextLen>0 do begin
    p:=IndexWord(SearchInText^,SearchInTextLen,PWord(SearchForText)^);
    if p<0 then exit;
    inc(SearchInText,p);
    dec(SearchInTextLen,p);
    if SearchInTextLen<SearchForTextLen then exit;
    if CompareMem(SearchInText,SearchForText,SearchForTextLen * 2) then
      exit(SearchInText);
    inc(SearchInText);
    dec(SearchInTextLen);
  end;
end;

function UTF16Pos(const SearchForText, SearchInText: UnicodeString; StartPos: SizeInt = 1): PtrInt;
// returns the character index, where the SearchForText starts in SearchInText
// an optional StartPos can be given (in UTF-16 codepoints, not in word)
// returns 0 if not found
var
  i: SizeInt;
  p: PUnicodeChar;
  StartPosP: PUnicodeChar;
begin
  Result:=0;
  if StartPos=1 then
  begin
    i:=System.Pos(SearchForText,SearchInText);
    if i>0 then
      Result:=UTF16Length(PUnicodeChar(SearchInText),i-1)+1;
  end
  else if StartPos>1 then
  begin
    // skip
    StartPosP:=UTF16CharStart(PUnicodeChar(SearchInText),Length(SearchInText),StartPos-1);
    if StartPosP=nil then exit;
    // search
    p:=UTF16PosP(PUnicodeChar(SearchForText),length(SearchForText),
                StartPosP,length(SearchInText)+PUnicodeChar(SearchInText)-StartPosP);
    // get UTF-8 position
    if p=nil then exit;
    Result:=StartPos+UTF16Length(StartPosP,p-StartPosP);
  end;
end;

function UTF16IsCombining(const AChar: PUnicodeChar): Boolean;
var
  ch: UnicodeChar;
begin
  ch := AChar[0];
  Result :=           // Combining Diacritical Marks (belongs to previos char)
    ( (ch >= #$300) and (ch <= #$36F) ) or    // 0300-036F
    ( (ch >= #$610) and (ch <= #$61A) ) or    // Arabic 0610..061A
    ( (ch >= #$64B) and (ch <= #$65F) ) or    // Arabic 064B..065F
    (  ch = #$670) or                         // Arabic 0670
    ( (ch >= #$6D6) and (ch <= #$6DC) ) or    // Arabic 06D6..06DC
    ( (ch >= #$6DF) and (ch <= #$6E4) ) or    // Arabic 06DF..06E4
    ( (ch >= #$6E7) and (ch <= #$6E8) ) or    // Arabic 06E7..06E8
    ( (ch >= #$6EA) and (ch <= #$6ED) ) or    // Arabic 06EA..06ED
    ( (ch >= #$8E4) and (ch <= #$8FE) ) or    // Arabic 08E4..08FE
    ( (ch >= #$1DC0) and (ch <= #$1DFF) ) or  // Combining Diacritical Marks Supplement 1DC0-1DFF
    ( (ch >= #$20D0) and (ch <= #$20FF) ) or  // Combining Diacritical Marks for Symbols 20D0-20FF
    ( (ch >= #$FE20) and (ch <= #$FE2F) );    // Combining half Marks FE20-FE2F
end;

{$ELSE}

function UTF8IsCombining(const AChar: PChar): Boolean;
begin
  Result :=
   ( (AChar[0] = #$CC) ) or                   // Combining Diacritical Marks (belongs to previos char) 0300-036F
   ( (AChar[0] = #$CD) and (AChar[1] in [#$80..#$AF]) ) or                        // Combining Diacritical Marks
   ( (AChar[0] = #$D8) and (AChar[1] in [#$90..#$9A]) ) or                        // Arabic 0610 (d890)..061A (d89a)
   ( (AChar[0] = #$D9) and (AChar[1] in [#$8b..#$9f, #$B0]) ) or                  // Arabic 064B (d98b)..065F (d99f) // 0670 (d9b0)

   ( (AChar[0] = #$DB) and (AChar[1] in [#$96..#$9C, #$9F..#$A4, #$A7..#$A8, #$AA..#$AD]) ) or // Arabic 06D6 (db96)..  .. ..06ED (dbad)
   ( (AChar[0] = #$E0) and (AChar[1] = #$A3) and (AChar[2] in [#$A4..#$BE]) ) or  // Arabic 08E4 (e0a3a4) ..08FE (e0a3be)

   ( (AChar[0] = #$E1) and (AChar[1] = #$B7) ) or                                 // Combining Diacritical Marks Supplement 1DC0-1DFF (e1b780)
   ( (AChar[0] = #$E2) and (AChar[1] = #$83) and (AChar[2] in [#$90..#$FF]) ) or  // Combining Diacritical Marks for Symbols 20D0-20FF
   ( (AChar[0] = #$EF) and (AChar[1] = #$B8) and (AChar[2] in [#$A0..#$AF]) );    // Combining half Marks FE20-FE2F
end;

{$ENDIF}


//---

function CodePointCopy(const s: string; StartCharIndex, CharCount: PtrInt): string;
begin
  {$IFDEF FPC_UNICODESTRINGS}
  Result := UnicodeCopy(s, StartCharIndex, CharCount);
  {$ELSE}
  Result := UTF8Copy(s, StartCharIndex, CharCount);
  {$ENDIF}
end;

function CodePointLength(const s: string): PtrInt;
begin
  {$IFDEF FPC_UNICODESTRINGS}
  Result := UTF16Length(s);
  {$ELSE}
  Result := UTF8Length(s);
  {$ENDIF}
end;

function CodePointPos(const SearchForText, SearchInText: string; StartPos: SizeInt = 1): PtrInt;
begin
  {$IFDEF FPC_UNICODESTRINGS}
  Result := UTF16Pos(SearchForText, SearchInText, StartPos);
  {$ELSE}
  Result := UTF8Pos(SearchForText, SearchInText, StartPos);
  {$ENDIF}
end;

function CodePointToWinCP(const s: string): AnsiString;
begin
  {$IFDEF FPC_UNICODESTRINGS}
  // ToDo: Don't convert through UTF-8.
  Result := UTF8ToWinCP(UTF16ToUTF8(s));
  {$ELSE}
  Result := UTF8ToWinCP(s);
  {$ENDIF}
end;

function WinCPToCodePoint(const s: AnsiString): string;
begin
  {$IFDEF FPC_UNICODESTRINGS}
  // ToDo: Don't convert through UTF-8.
  Result := UTF8ToUTF16(WinCPToUTF8(s));
  {$ELSE}
  Result := WinCPToUTF8(s);
  {$ENDIF}
end;

function CodeUnitCount(p: PChar): integer;
begin
  {$IFDEF FPC_UNICODESTRINGS}
  if TCharacter.IsHighSurrogate(p^) then
    Result := 2
  else
    Result := 1
  {$ELSE}
  Result := UTF8CharacterLength(p);
  {$ENDIF}
end;

function IsCombining(const AChar: PChar): Boolean;
begin
  {$IFDEF FPC_UNICODESTRINGS}
  Result := UTF16IsCombining(AChar);
  {$ELSE}
  Result := UTF8IsCombining(AChar);
  {$ENDIF}
end;

{ TUnicodeEnumeratorBase }

constructor TUnicodeEnumeratorBase.Create(const A: String);
begin
  fCurrentPos := PChar(A); // Note: if A='' then PChar(A) returns a pointer to a #0 string
  fEndPos := fCurrentPos + length(A);
end;

{ TCodePointEnumerator }

function TCodePointEnumerator.MoveNext: Boolean;
begin
  if fCurrentPos < fEndPos then
  begin
    fCurrentCodeUnitCount := CodeUnitCount(fCurrentPos);
    SetLength(fCurrent, fCurrentCodeUnitCount);
    Move(fCurrentPos^, fCurrent[1], fCurrentCodeUnitCount*SizeOf(Char));
    inc(fCurrentPos, fCurrentCodeUnitCount);
    Result := true;
  end else
    Result := false;
end;

{ TUnicodeCharacterEnumerator }

function TUnicodeCharacterEnumerator.MoveNext: Boolean;
var
  NextCP: PChar;
  NextCUCount: Integer;
begin
  if fCurrentPos < fEndPos then
  begin
    fCurrentCodePointCount := 0;
    NextCP := fCurrentPos;
    repeat
      NextCUCount := CodeUnitCount(NextCP); // Prepare for combining diacritical marks.
      NextCP += NextCUCount; // Prepare for combining diacritical marks.
      Inc(fCurrentCodePointCount);
    until not IsCombining(NextCP);
    fCurrentCodeUnitCount := NextCP - fCurrentPos;  // Pointer arithmetics.
    SetLength(fCurrent, fCurrentCodeUnitCount);
    Move(fCurrentPos^, fCurrent[1], fCurrentCodeUnitCount*SizeOf(Char));
    inc(fCurrentPos, fCurrentCodeUnitCount);
    Result := true;
  end else
    Result := false;
end;

//---
// Enumerator
//---
operator Enumerator(A: String): TUnicodeCharacterEnumerator;
begin
  Result := TUnicodeCharacterEnumerator.Create(A);
end;

end.

